// |---------------------------------------------------------|
// |                                                         |
// |      Virtual Pascal Runtime Library.  Version 2.1.      |
// |      Test classes for numeric routines in Math and System
// |      All reference values are from XMaxima v5.9.0       |
// |      ---------------------------------------------------|
// |      Copyright (C) 2003 vpascal.com                     |
// |---------------------------------------------------------|

unit MathTest;

interface

uses
  Classes, SysUtils, UnitTest, Math;

type
  TTestRecord = record
    fX      : Extended;
    fExpect : Extended;
  end;

  tMathTest = class(tTest)
  private
    function TestSpecial(X: Extended; _CWMask: SmallWord;
      _Except: ExceptClass; _Expect: Extended; const _S: String): Boolean;
  protected
    function MaxDelta: Extended; virtual;
    function GetError(_Value1, _Value2: Extended): Extended;
    function TestSimilar(_Value1, _Value2: Extended; const _S: String): Boolean;
    function ValidateList( _List: array of TTestRecord; const _S: String ): Boolean;
    function TestInf(X: Extended; const _S: String ): Boolean;
    function TestFunc(X: Extended): Extended; virtual; abstract;
  end;

  tTestSinCos = class(tMathTest)
  protected
    function TestSinCos(X, SinX, CosX: Extended; const _S: String): Boolean;
  public
    function Run: Boolean; override;
  end;

  tTestSinCosDeriv = class(tMathTest)
  public
    function Run: Boolean; override;
  end;

  tTestTan = class(tMathTest)
    function TestFunc(X: Extended): Extended; override;
    function Run: Boolean; override;
  end;

  tTestArcTan = class(tMathTest)
    function TestFunc(X: Extended): Extended; override;
    function Run: Boolean; override;
  end;

  tTestIntPower = class(tMathTest)
  public
    function Run: Boolean; override;
  end;

  tTestPower = class(tMathTest)
  public
    function Run: Boolean; override;
  end;

  tTestArctanh = class(tMathTest)
  protected
    function TestFunc(X: Extended): Extended; override;
  public
    function Run: Boolean; override;
  end;

  tTestRounding = class(tMathTest)
  public
    function Run: Boolean; override;
  end;

  tTestHypot = class(tMathTest)
  public
    function Run: Boolean; override;
  end;

  // Exponential function test
  tTestExp = class(tMathTest)
  protected
    function MaxDelta: Extended; override;
    function TestFunc(X: Extended): Extended; override;
  public
    function Run: Boolean; override;
  end;

  // All of these functions rely on Exp() and allow slightly more error
  tTestSinh = class(tTestExp)
    function TestFunc(X: Extended): Extended; override;
    function Run: Boolean; override;
  end;

  tTestCosh = class(tTestExp)
    function TestFunc(X: Extended): Extended; override;
    function Run: Boolean; override;
  end;

  tTestTanh = class(tTestExp)
    function TestFunc(X: Extended): Extended; override;
    function Run: Boolean; override;
  end;

  tTestCubeRoot = class(tTestExp)
    function TestFunc(X: Extended): Extended; override;
    function Run: Boolean; override;
  end;



implementation

// tMathTest

function tMathTest.MaxDelta: Extended;
begin
  Result := 1e-17; // For most math tests, this is the accuracy we need
end;

function tMathTest.GetError(_Value1, _Value2: Extended): Extended;
begin
  if _Value2 = 0 then
    Result := Abs(_Value1)
  else if _Value1 = 0 then
    Result := Abs(_Value2)
  else
    Result := Abs(_Value1/_Value2-1);
end;

function tMathTest.TestSimilar(_Value1, _Value2: Extended; const _S: String): Boolean;
begin
  Result := Test( GetError(_Value1, _Value2) < MaxDelta, _S )
end;

function tMathTest.ValidateList( _List: array of TTestRecord; const _S: String ): Boolean;
var
  i: Integer;
begin
  Result := True;
  for i := Low(_List) to High(_List) do
    Result := Result and TestSimilar( TestFunc(_List[i].fX), _List[i].fExpect, Format(_S, [_List[i].fX]) );
end;

// Verify a special case
function tMathTest.TestSpecial(X: Extended; _CWMask: SmallWord;
  _Except: ExceptClass; _Expect: Extended; const _S: String): Boolean;
var
  cw: SmallWord;
  Y: Extended;
begin
  Result := False;
  cw := Get87Cw;
  // By default, overflow causes an exception
  try
    Y := TestFunc(X);
  except
    on E:Exception do
      if E is _Except then
        Result := True
      else
        // Not the right exception
        FailReason := Format( '%s (Except = %s)', [_S, E.Message] );
  end;
  if not Result then
    Exit;

  // Clear the relevant exception
  Set87CW(CW or _CWMask);
  try
    try
      Y := TestFunc(X);
      Result := Y = _Expect;
      if not Result then
        FailReason := Format('%s: %g <> %g', [_S, Y, _Expect]);
    except
      on E:Exception do
        // We don't want any exceptions
        FailReason := Format( '%s (!!Except = %s)', [_S, E.Message] );
    end;
  finally
    // Restore the control word
    Set87CW(CW);
  end;
end;

// Verify that the function can return INF or raise an exception
function tMathTest.TestInf(X: Extended; const _S: String ): Boolean;
begin
{$IFDEF LINUX}
  Result := TestSpecial(X, FPUCW_OM, EFloatError, PosINF, _S);
{$ELSE}
  Result := TestSpecial(X, FPUCW_OM, EOverflow, PosINF, _S);
{$ENDIF}
end;

// Test Sine, Cosine, reverse and related functions

function tTestSinCos.TestSinCos(X, SinX, CosX: Extended; const _S: String): Boolean;
var
  MySinX: Extended;
  MyCosX: Extended;
begin
  SinCos(X, MySinX, MyCosX);
  Result := TestSimilar(SinX, MySinX, _S + ' (sin)') and
            TestSimilar(CosX, MyCosX, _S + ' (cos)');
end;

function tTestSinCos.Run: Boolean;
begin
  // Test sin
  Result := TestSimilar(sin(2*pi), 0, 'sin(2*pi)');
  Result := Result and TestSimilar(sin(pi), 0, 'sin(pi)');
  Result := Result and TestSimilar(sin(pi/2), 1, 'sin(pi)');
  Result := Result and TestSimilar(sin(-pi/2), -1, 'sin(pi)');
  Result := Result and TestSimilar(sin(10), -0.5440211108893698134, 'sin(10)');

  // Test SinCos
  Result := Result and TestSinCos(0,      0,  1, 'SinCos(0)');
  Result := Result and TestSinCos(pi,     0, -1, 'SinCos(pi)');
  Result := Result and TestSinCos(pi/2,   1,  0, 'SinCos(pi/2)');
  Result := Result and TestSinCos(-pi/2, -1,  0, 'SinCos(-pi/2)');
  Result := Result and TestSinCos(-pi,    0, -1, 'SinCos(-pi)');
  Result := Result and TestSinCos(10,    -0.5440211108893698134, -0.8390715290764524522588639, 'SinCos(10)');

  // Test ArcSin
  Result := Result and TestSimilar(ArcSin(0), 0, 'Arcsin(0)');
  Result := Result and TestSimilar(ArcSin(1), pi/2, 'ArcSin(1)');
  Result := Result and TestSimilar(ArcSin(-1), -pi/2, 'ArcSin(-1)');
  Result := Result and TestSimilar(ArcSin(0.5440211108893698134), 0.5752220392306202846, 'ArcSin(-0.5440211108893698134)');

  // Test Cos
  Result := Result and TestSimilar(cos(2*pi), 1, 'cos(2*pi)');
  Result := Result and TestSimilar(cos(pi), -1, 'cos(pi)');
  Result := Result and TestSimilar(cos(pi/2), 0, 'cos(pi)');
  Result := Result and TestSimilar(cos(-pi/2), 0, 'cos(pi)');
  Result := Result and TestSimilar(cos(10), -0.8390715290764524522588639, 'cos(10)');

  // Test ArcCos
  Result := Result and TestSimilar(ArcCos(0), pi/2, 'ArcCos(0)');
  Result := Result and TestSimilar(ArcCos(1), 0, 'ArcCos(1)');
  Result := Result and TestSimilar(ArcCos(-1), pi, 'ArcCos(-1)');
  Result := Result and TestSimilar(ArcCos(0.8390715290764524522588639), 0.5752220392306202846, 'ArcSin(-0.5440211108893698134)');
end;

// Test derivatives of Sin and Cos

function tTestSinCosDeriv.Run: Boolean;
begin
  // Test Sinc = sin(x)/x
  Result := TestSimilar(Sinc(pi/2),    2/pi, 'Sinc(pi/2)');
  Result := Result and TestSimilar(Sinc(0.1),     0.9983341664682815230681419841, 'Sinc(0.1)');
  Result := Result and TestSimilar(Sinc(0.01),    0.9999833334166664682542438269, 'Sinc(0.01)');
  Result := Result and TestSimilar(Sinc(0.0001),  0.9999999983333333341666666665, 'Sinc(0.0001)');
  Result := Result and TestSimilar(Sinc(0.00001), 0.9999999999833333333334166666, 'Sinc(0.00001)');

  // Test Secant = 1/cos(x)
  Result := Result and TestSimilar(Secant(0.1),     1.005020918400455428465114101311635819397, 'Secant(0.1)');
  Result := Result and TestSimilar(Secant(0.0001),  1.000000005000000020833333418055555899058, 'Secant(0.0001)');
  Result := Result and TestSimilar(Secant(1.5),     14.13683290296990308192322843418549445144, 'Secant(1.5)');

  // Test CoSecant = 1/sin(x)
  Result := Result and TestSimilar(CoSecant(0.1),   10.0166861316347766487063525420806837369, 'CoSecant(0.1)');
  Result := Result and TestSimilar(CoSecant(0.001), 1000.000166666686111113161375871362455198, 'CoSecant(0.001)');
  Result := Result and TestSimilar(CoSecant(1.5),   1.002511304246724909954183649508739717262, 'CoSecant(1.5)');
end;

// Test hyperbolic sine

function tTestSinh.TestFunc(X: Extended): Extended;
begin
  Result := Sinh(X);
end;

function tTestSinh.Run: Boolean;
const
  TESTS = 10;
const
  TestValues: array[1..TESTS] of TTestRecord = (
    (fX: 0;        fExpect:  0.0),
    (fX: 1e-5;     fExpect:  1.000000000016666666666750000000000199E-5),
    (fX: 0.01;     fExpect:  1.000016666750000198412973986141173801E-2),
    (fX: 0.43;     fExpect:  0.44337421441248242854091898371325716501),
    (fX: 0.5;      fExpect:  0.52109530549374736162242562641149155910),
    (fX: -0.43;    fExpect: -0.44337421441248242854091898371325716501),
    (fX: 1.07;     fExpect:  1.28618549127914500287233264178838866631),
    (fX: 37.81;    fExpect:  1.3171776868410822017570207570227961E16),
    (fX: 500;      fExpect:  7.0179610892641870536988516642045604E216),
    (fX: -1.07;    fExpect: -1.28618549127914500287233264178838866631)
  );

begin
  Result := ValidateList( TestValues, 'Hyperbolic sine of %.3f' );
  // Verify that a too large value works correctly
  Result := Result and TestInf( 15000, 'Hyperbolic sine' );
end;

// Test hyperbolic cosine

function tTestCosh.TestFunc(X: Extended): Extended;
begin
  Result := Cosh(X);
end;

function tTestCosh.Run: Boolean;
const
  TESTS = 8;
const
  TestValues: array[1..TESTS] of TTestRecord = (
    (fX: 0;        fExpect:  1.0),
    (fX: 1e-5;     fExpect:  1.000000000050000000000416666666668055556),
    (fX: 0.5;      fExpect:  1.127625965206380785226225161402672012548),
    (fX: -0.43;    fExpect:  1.093883309135798973159934482201158681244),
    (fX: 1.07;     fExpect:  1.629194008697851671005538131187143538581),
    (fX: 37.81;    fExpect:  1.317177686841082201757020757022799942863E16),
    (fX: 500;      fExpect:  7.017961089264187053698851664204560410903E216),
    (fX: -1.07;    fExpect:  1.629194008697851671005538131187143538581)
  );

begin
  Result := ValidateList( TestValues, 'Hyperbolic cos of %.3f' );
  // Verify that a too large value works correctly
  Result := Result and TestInf( 15000, 'Hyperbolic cos' );
end;

// Test hyperbolic Tan

function tTestTanh.TestFunc(X: Extended): Extended;
begin
  Result := Tanh(X);
end;

function tTestTanh.Run: Boolean;
const
  TESTS = 7;
const
  TestValues: array[1..TESTS] of TTestRecord = (
    (fX: 0;        fExpect:  0.0),
    (fX: 1e-5;     fExpect:  9.99999999966666666667999999999946032559E-6),
    (fX: 0.11;     fExpect:  1.095584702144295290839537110784533345514E-1),
    (fX: -0.11;    fExpect: -1.095584702144295290839537110784533345514E-1),
    (fX: 1;        fExpect:  7.615941559557648881194582826047935904127E-1),
    (fX: 37.81;    fExpect:  9.999999999999999999999999999999971180843E-1),
    (fX: 500;      fExpect:  1)
  );

begin
  Result := ValidateList( TestValues, 'Hyperbolic tan of %.3f' );
end;

// Test Tangent

function tTestTan.TestFunc(X: Extended): Extended;
begin
  Result := Tan(X);
end;

function tTestTan.Run: Boolean;
const
  TESTS = 5;
const
  TestValues: array[1..TESTS] of TTestRecord = (
    (fX: 0;      fExpect:  0.0),
    (fX: pi;     fExpect:  0.0),
    (fX: pi/4;   fExpect:  1.0),
    (fX: 10;     fExpect:  6.483608274590866712591249330098086768169E-1),
    (fX: -10;    fExpect: -6.483608274590866712591249330098086768169E-1)
  );

begin
  Result := ValidateList( TestValues, 'Tan of %.3f');
end;

// Test ArcTan

function tTestArcTan.TestFunc(X: Extended): Extended;
begin
  Result := ArcTan(X);
end;

function tTestArcTan.Run: Boolean;
const
  TESTS = 2;
const
  TestValues: array[1..TESTS] of TTestRecord = (
    (fX: 0;      fExpect:  0.0),
    (fX: 1;      fExpect:  pi/4)
  );

begin
  Result := ValidateList( TestValues, 'ArcTan of %.3f');

  // Test special values
  Result := Result and TestSimilar(ArcTan(PosInf), pi/2, 'ArcTan(pi/2)');
  Result := Result and TestSimilar(ArcTan(-PosInf), -pi/2, 'ArcTan(-pi/2)');
end;

// Test X^N

function tTestIntPower.Run: Boolean;
var
  x: Extended;
begin
  x := 3.78;
  Result := Test( IntPower(x, 1) = x, 'Power of 1' );
  Result := Result and Test( IntPower(x, 0) = 1, 'Power of 0' );
  Result := Result and TestSimilar( IntPower(x,  6), 2917.096519063104, 'Power of 6' );
  Result := Result and TestSimilar( IntPower(x, -6), 3.42806620715167200E-4, 'Power of -6' );
  Result := Result and TestSimilar( IntPower(x, 11), 2.25117780463758685E+6, 'Power of 11' );

  x := -781.992;
  Result := Result and Test( IntPower(x, 1) = x, '< 0 Power of 1' );
  Result := Result and Test( IntPower(x, 0) = 1, '< 0 Power of 0' );
  Result := Result and TestSimilar( IntPower(x,  6),  2.28672458389969773E+17, '< 0 Power of 6' );
  Result := Result and TestSimilar( IntPower(x, -6),  4.37306708049045427E-18, '< 0 Power of -6' );
  Result := Result and TestSimilar( IntPower(x, 11), -6.68690897427498711E+31, '< 0 Power of 11' );
end;

// Test X^Y

function tTestPower.Run: Boolean;
var
  x: Extended;
begin
  x := 1;
  Result := Test( Power(x, 8) = 1, '1^N' );
end;

// Test arctanh

function tTestArctanh.TestFunc(X: Extended): Extended;
begin
  Result := Arctanh(X);
end;

function tTestArctanh.Run: Boolean;
const
  TESTS = 4;
const
  TestValues: array[1..TESTS] of TTestRecord = (
    (fX: 0;        fExpect:  0.0),
    (fX: 0.9;      fExpect:  1.4722194895832202300045137),
    (fX: 0.3;      fExpect:  0.3095196042031117154740673),
    (fX: -0.99;    fExpect: -2.6466524123622461977050606)
  );

begin
  Result := ValidateList( TestValues, 'Arctanh of %.3f' );

  // Infinites have to be done by hand
  Result := Result and Test( Arctanh(-1) = NegINF, 'Arctanh(-1)' );
  Result := Result and Test( Arctanh(1)  = PosINF, 'Arctanh(1)' );
  Result := Result and Test( Arctanh(2)  = PosINF, 'Arctanh(2)' );
end;

// tTestExp - test Exp() function

function tTestExp.MaxDelta: Extended;
begin
  Result := 5e-17; // For exp() tests, we have to allow a little more error
end;

function tTestExp.TestFunc(X: Extended): Extended;
begin
  Result := Exp(X);
end;

function tTestExp.Run: Boolean;
const
  TESTS = 11;
const
  TestValues: array[1..TESTS] of TTestRecord = (
    (fX: 0;        fExpect: 1.0),
    (fX: 15.33;    fExpect: 4.547098976503116215515867881787616E6   ),
    (fX: 86;       fExpect: 2.2352466037347150474430657323327147E37 ),
    (fX: 86.971;   fExpect: 5.9023557992169759618854720233041498E37 ),
    (fX: 500;      fExpect: 1.403592217852837410739770332840912E217 ),
    (fX: 901.7;    fExpect: 4.011754349665291699239689171416141E391 ),
    (fX: 1198.8;   fExpect: 4.287705078619925558243032864045091E520 ),
    (fX: 4000;     fExpect: 1.506355970050524900975919074741499E1737),
    (fX: -4021.99; fExpect: 1.870409630456545314940621142269705E-1747),
    (fX: 8729.71;  fExpect: 1.840270293682485180295346690002773E3791),
    (fX: -1198.71; fExpect: 2.551887929888568476122029564819985E-521)
  );

begin
  Result := ValidateList( TestValues, 'Power of %.3f' );

  // Verify that a too large value works correctly
  Result := Result and TestInf( 15000, 'Exp' );
end;

// Test round control

function tTestRounding.Run: Boolean;
var
  CW: SmallWord;
  RC: tRoundControl;
  x: array[tRoundControl] of Extended;
begin
  CW := Get87CW;
  RC := GetRoundControl;
  Result := Test( RC = rcNearest, 'Default round control setting' );
  for RC := Low(x) to High(x) do
    begin
      SetRoundControl(RC);
      x[RC] := Power( 3.71, 1.94 );
    end;
  Result := Result and Test( x[rcDown] < x[rcUp], 'Round down < Round up' );
  Result := Result and Test( x[rcChop] < x[rcUp], 'Chop < Round up' );
  Result := Result and TestSimilar( x[rcDown], x[rcUp], 'Round down =~ Round up' );
  Result := Result and TestSimilar( x[rcDown], x[rcNearest], 'Round down =~ nearest' );
  Result := Result and TestSimilar( x[rcDown], x[rcChop], 'Round down =~ chop' );

  // Test that we can set and restore the control word
  Set87CW(0);
  Set87CW(CW);
  Result := Result and Test(Get87CW = CW, 'Get/Set Control Flag');
end;

// Test hypotenuse function

function tTestHypot.Run: Boolean;
var
  x: Extended;
begin
  x := 17.81;
  Result := Test( Hypot( X, 0 ) = X, 'Hypot(x,0) = x') ;
  Result := Result and Test( Hypot( 0, X ) = X, 'Hypot(0,x) = x') ;
  Result := Result and TestSimilar( Hypot(  x,  9.67 ), 20.26585798825206413323, 'Hypot(x,y)' );
  Result := Result and TestSimilar( Hypot(  x, -9.67 ), 20.26585798825206413323, 'Hypot(x,y)' );
  Result := Result and TestSimilar( Hypot( -x,  9.67 ), 20.26585798825206413323, 'Hypot(x,y)' );
end;

// Test cubic root function

function tTestCubeRoot.TestFunc(X: Extended): Extended;
begin
  Result := CubeRoot(x);
end;

function tTestCubeRoot.Run: Boolean;
const
  TESTS = 5;
const
  TestValues: array[1..TESTS] of TTestRecord = (
    (fX: 0;        fExpect:  0.0),
    (fX: 17.89;    fExpect:  2.615391935034079367134128496586401686476),
    (fX: -17.89;   fExpect: -2.615391935034079367134128496586401686476),
    (fX: 178831.01;fExpect:  56.33966709437273912752862213176555814641),
    (fX: 125E899;  fExpect:  2.320794416806389446205038175459723288E300)
  );

begin
  Result := ValidateList( TestValues, 'Cuberoot of %.3f' );
end;

initialization
  RegisterTests( [tTestSinCos, tTestSinCosDeriv, tTestTan, tTestArcTan,
    tTestIntPower, tTestPower, tTestExp, tTestSinh, tTestCosh, tTestTanh,
    tTestArctanh, tTestRounding, tTestHypot, tTestCubeRoot] );
end.


