// |---------------------------------------------------------|
// |                                                         |
// |      Virtual Pascal Runtime Library.  Version 2.1.      |
// |      Various classes used for structured unit testing   |
// |      ---------------------------------------------------|
// |      Copyright (C) 2003 vpascal.com                     |
// |---------------------------------------------------------|

unit UnitTest;
{&Use32+,G5+}

interface

uses
  SysUtils, Classes;

type
  tTestClass = class of tTest;
  tTest = class
  private
    fFailReason: String;
    fStartTicks: Comp;
  protected
    function Test(_Ok: Boolean; const _Reason: String): Boolean; virtual;
    procedure StartTimer;
    function TimerTicks: Comp;
  public
    function Run: Boolean; virtual; abstract;
    function Name: String; virtual;
    property FailReason: String read fFailReason write fFailReason;
  end;

type
  tTestList = class(TList)
  private
    fFailures: TStringList;
    function RunTest(_Test: tTest): Boolean;
  public
    constructor Create;
    function Run: Boolean;
    property Failures: TStringList read fFailures;
  end;

const
  TheTests: tTestList = nil;

procedure RegisterTest( _Test: tTestClass );
procedure RegisterTests( _Tests: array of tTestClass );
function RunTests: Boolean;

implementation

procedure RegisterTest( _Test: tTestClass );
begin
  if not assigned(TheTests) then
    TheTests := tTestList.Create;
  TheTests.Add(_Test);
end;

procedure RegisterTests( _Tests: array of tTestClass );
var
  i: Integer;
begin
  for i := Low(_Tests) to High(_Tests) do
    RegisterTest( _Tests[i] );
end;

function RunTests: Boolean;
begin
  if assigned(TheTests) then
    Result := TheTests.Run
  else
    Result := True;
end;

// --- Helper routine for getting timer ticks quickly ---

function GetCPUTicks: Comp; assembler; {&frame-} {&uses ecx,edx}
asm
  push   0
  push   0
  mov    ecx,esp
  rdtsc
  mov    [ecx+4],edx
  mov    [ecx],eax
  fild   qword ptr [ecx]
  pop    eax
  pop    eax
end;


// --- tTest ---

function tTest.Name: String;
begin
  Result := ClassName; // If not overriden, this is a good choice
end;

function tTest.Test(_Ok: Boolean; const _Reason: String): Boolean;
begin
  Result := _Ok;
  if not _Ok then
    FailReason := _Reason;
end;

procedure tTest.StartTimer;
begin
  fStartTicks := GetCPUTicks;
end;

function tTest.TimerTicks: Comp;
begin
  Result := GetCPUTicks - fStartTicks;
end;

// --- tTestList ---

constructor tTestList.Create;
begin
  inherited Create;
  fFailures := tStringList.Create;
end;

// Run a single test and return the result.  Verify that
// registers are not modified and that no memory leaks exist
function tTestList.RunTest(_Test: tTest): Boolean;
var
  saveEbx, saveEsi, saveEdi, saveEbp: Longint;
  postEbx, postEsi, postEdi, postEbp: Longint;
  HStatus: THeapStatus;
  StartingMem: Longint;
  Leak: Longint;
begin
  HStatus := GetHeapStatus;
  StartingMem := HStatus.TotalAllocated;
  try
    asm
      // Prevents compiler from allocating these registers for use
      {&Alters ebx,esi,edi,ebp}
      mov   saveEbx,ebx
      mov   saveEsi,esi
      mov   saveEdi,edi
      mov   saveEbp,ebp
    end;
    Result := _Test.Run;
    asm
      {&Alters ebx,esi,edi,ebp}
      mov   postEbx,ebx
      mov   postEsi,esi
      mov   postEdi,edi
      mov   postEbp,ebp
    end;
  except
    // Catch any escaping exceptions!
    on E:Exception do
      Result := _Test.Test(False, Format('Unexpected exception %s (%s)', [E.Message, _Test.FailReason]));
  end;
  Result := Result and _Test.Test(saveEbx = postEbx, 'ebx register changed');
  Result := Result and _Test.Test(saveEsi = postEsi, 'esi register changed');
  Result := Result and _Test.Test(saveEdi = postEdi, 'edi register changed');
  Result := Result and _Test.Test(saveEbp = postEbp, 'ebp register changed');
  HStatus := GetHeapStatus;
  Leak := HStatus.TotalAllocated-StartingMem;
  // Do not fail on memory leak.  We need to add a warning type.
  if Leak <> 0 then
    Writeln(' !! Memory leak in bytes: ',Leak);
  // Result := Result and _Test.Test(Leak = 0, Format('Memory leak of %d bytes',[Leak]) );
end;

function tTestList.Run: Boolean;
var
  i: Integer;
  rc: Boolean;
  Test: tTest;
  TestClass: tTestClass;
begin
  Result := True;
  for i := 0 to pred(Count) do
    begin
      TestClass := tTestClass(Items[i]);
      Test := TestClass.Create;
      Write( ' - ',Test.Name );
      rc := RunTest(Test);
      if rc then
        writeln(' - Ok')
      else
      begin
        Writeln(' - Failed ('+Test.FailReason,')');
        Failures.Add( Format('%s: %s', [Test.Name, Test.FailReason]) );
      end;
      Result := Result and rc;
      Test.Free;
    end;
end;

end.

