// |---------------------------------------------------------|
// |                                                         |
// |      Virtual Pascal Runtime Library.  Version 2.1.      |
// |      Unit Test for Exception Handling mechanism         |
// |      ---------------------------------------------------|
// |      Copyright (C) 2003 vpascal.com                     |
// |---------------------------------------------------------|

// NOTE: If running this test case from the IDE, make sure all
//       exceptions are handled by the program and not the IDE!
//       To do this, go to View->Exceptions and click None.

unit XcptTest;
{&Use32+,S-,R-,Q-}

interface

uses
  Classes, SysUtils, UnitTest, Math;

type
  tTestMethod = function: Boolean of object;

  tExceptionTest = class(tTest)
  protected
    function Test(_Routine: tTestMethod; const _Testing: String): Boolean;
  end;

  tTestBasicExceptions = class(tExceptionTest)
  protected
    function Basic: Boolean;
    function Specific: Boolean;
    function TryFinally: Boolean;
  public
    function Run: Boolean; override;
  end;

  tTestOSExceptions = class(tExceptionTest)
  protected
    function IntDivZero: Boolean;
    function ArrayBounds: Boolean;
    function IntOverflow: Boolean;
    function FloatDivZero: Boolean;
    function FloatOverflow: Boolean;
    function StackOverflow: Boolean;
    function Denormal: Boolean;
    function Access: Boolean;
    function Privilege: Boolean;
  public
    function Run: Boolean; override;
  end;

implementation

// Test wrapper for exception tests to catch unexpected exceptions

function tExceptionTest.Test(_Routine: tTestMethod; const _Testing: String): Boolean;
begin
  try
    Result := _Routine;
    Result := inherited Test(Result, _Testing);
  except
    on E:Exception do
      Result := inherited Test(False, Format('%s: Unexpected exception "%s"', [_Testing, E.Message]));
  end;
end;

// Test of basic exception handling mechanism

function tTestBasicExceptions.Basic: Boolean;
begin
  // The simplest possible exception
  Result := False;
  try
    raise Exception.Create('Exception');
  except
    Result := True;
  end;
end;

function tTestBasicExceptions.Specific: Boolean;
begin
  // Catch a specific exception
  Result := False;
  try
    raise EOutOfMemory.Create('Test memory Exception');
  except
    on E:EOutOfMemory do
      Result := True;
  end;
end;

function tTestBasicExceptions.TryFinally: Boolean;
begin
  // Try..finally block
  Result := False;
  try
    try
      raise EOutOfMemory.Create('Bob');
    finally
      Result := True;
    end;
  except
    on EOutOfMemory do
      ; // Nothing
    on Exception do
      Result := False; // Should not get here
  end;
end;

function tTestBasicExceptions.Run: Boolean;
begin
  Result := Test(Basic, 'Basic');
  Result := Result and Test(Specific, 'Specific');
  Result := Result and Test(TryFinally, 'Try..Finally');
end;

// Test mapping of OS exceptions to language exceptions

function tTestOSExceptions.IntDivZero: Boolean;
var
  N: Integer;
begin
  Result := False;
  try
    N := 0;
    N := 17 div N;
  except
    on EDivByZero do
      Result := True;
  end;
end;

function tTestOSExceptions.ArrayBounds: Boolean;
var
  N: Integer;
  X: Array[7..17] of Integer;
begin
  Result := False;
  {$R-}
  try
    N := 18;
    N := X[N]; // Error, but ignored
  except
    exit; // No exception should be raised
  end;

  {$R+} // Compiler must generate range checking code
  try
    N := 18;
    N := X[N]; // Range check error
  except
    on ERangeError do
      Result := True;
  end;
end;

function tTestOSExceptions.IntOverflow: Boolean;
var
  N: Longint;
begin
  Result := False;
  {$Q-}
  try
    N := $70000000;
    N := N*N;  // Silent overflow in Q-
  except
    exit; // Error
  end;

  {$Q+} // Compiler must generate overflow checking code
  try
    N := $70000000;
    N := N*N;  // Integer overflow
  except
    on EIntOverflow do
      Result := True;
  end;
end;

function tTestOSExceptions.FloatDivZero: Boolean;
var
  N: Integer;
  X: Double;
begin
  Result := False;
  try
    N := 0;
    X := 17 / N;
  except
    on EZeroDivide do
      Result := True;
{$IFDEF LINUX}
    on EFloatError do
      Result := True; // Linux does not raise specific float exceptions
{$ENDIF}
  end;
end;

function tTestOSExceptions.FloatOverflow: Boolean;
var
  X: Double;
begin
  Result := False;
  {$Q-}
  try
    X := MaxDouble/2;
    X := X*10; // Overflow, ignored in Q-
  except
    exit; // Error
  end;

  {$Q+} // Compiler must generate overflow checking code
  try
    X := MaxDouble/2;
    X := X*10; // Overflow
  except
    on EOverflow do
      Result := True;
{$IFDEF LINUX}
    on EFloatError do
      Result := True; // Linux does not raise specific float exceptions
{$ENDIF}
  end;
end;

{$IFNDEF LINUX}
// We only catch this if stack checking is enabled
// Stack checking does not work in Linux
{$S-}
{$ENDIF}
function tTestOSExceptions.StackOverflow: Boolean;

  procedure Recursive;
  var
    s: Array[1..1024] of byte; // Eat some stack
  begin
    Recursive;
  end;

begin
  Result := False;
  try
    Recursive
  except
    on EStackOverflow do
      Result := True;
  end;
end;

function tTestOSExceptions.Denormal: Boolean;
var
  X: Extended;
begin
  try
    X := FNan + 17 / FNan;
    Result := (X = FNan);
  except
    Result := False;
  end;
end;

function tTestOSExceptions.Access: Boolean;
var
  P: ^Longint;
  X: Longint;
begin
  Result := False;
  try
    P := nil;
    X := P^; // Access violation
  except
    on EAccessViolation do
      Result := True;
  end;
end;

function tTestOSExceptions.Privilege: Boolean;
begin
  Result := False;
  try
    asm
      mov   eax,cr0 // Privileged instruction cannot run outside kernel mode
    end;
  except
    on EPrivilege do
      Result := True;
{$IFDEF LINUX}
    on EAccessViolation do // Linux thinks this is an AV
      Result := True;
{$ENDIF}
  end;
end;

function tTestOSExceptions.Run: Boolean;
begin
  Result := Test(IntDivZero, 'Int divide by zero');
  Result := Result and Test(ArrayBounds, 'Array bounds');
  Result := Result and Test(IntOverflow, 'Int overflow');
  Result := Result and Test(FloatDivZero, 'Float divide by zero');
  Result := Result and Test(FloatOverflow, 'Float overflow');
  Result := Result and Test(Denormal, 'Denormal operand');
  Result := Result and Test(Access, 'Access violation');
  Result := Result and Test(Privilege, 'Privileged instruction');

{$IFNDEF LINUX}
  Result := Result and Test(StackOverflow, 'Stack overflow');
{$ENDIF}
end;

initialization
  RegisterTests( [tTestBasicExceptions, tTestOSExceptions] );
end.

