//
// This source file is included as a unit used by VpSysW32.Pas, and
// implements default keyboard handling for Win32 Virtual Pascal programs.
//
// By defining the conditional define KEYDLL, compiling this file results
// in a DLL, VpKbdW32.Dll to be produced.  When this DLL is present, the
// keyboard handling functions of the DLL supercede those statically
// linked into the program, allowing for advanced customization of the
// keyboard handler.

{$Ifdef KeyDll} library {$Else} unit {$Endif}
  VpKbdW32;

{&Delphi+,S-,R-,Optimize+,J+,X+}

{$Ifndef KeyDll}
interface
{$Endif}

uses
  Windows;

{$Ifndef KeyDll}
procedure KbdInit(var _pSysKeyCount, _pSysKeyQue, _pSysShiftState, _pSysMouCount, _pSysMouQue);
procedure KbdUpdateEventQueues;

implementation
{$Endif}

// Type definitions duplicated from VpSysLow
type
  PSysPoint = ^TSysPoint;
  TSysPoint = packed record
    X,Y: SmallInt;
  end;

  TSysMouseEvent = packed record
    smeTime:    Longint;
    smePos:     TSysPoint;
    smeButtons: Byte;
  end;

  TSysKeyEvent = packed record
    skeKeyCode:    SmallWord;
    skeShiftState: Byte;
  end;

  TSysMouQUeue = array[0..15] of TSysMouseEvent;
  PSysMouQueue = ^TSysMouQueue;
  TSysKeyQueue = array[0..15] of TSysKeyEvent;
  PSysKeyQueue = ^tSysKeyQueue;

// Variables holding the keyboard and mouse queues
var
  SysKeyQue: TSysKeyQueue;
  SysMouQue: TSysMouQueue;

const
  SysKeyCount  : Integer = 0;
  SysMouCount  : Integer = 0;
  SysShiftState: Byte = 0;
  SysPlatform  : Integer = 0;

// Initialisation - set VpSysW32 pointer variables to point to to here
procedure KbdInit(var _pSysKeyCount, _pSysKeyQue, _pSysShiftState, _pSysMouCount, _pSysMouQue);
var
  OSVersionInfo: TOSVersionInfo;
begin
  // Link VpSysW32 variables to the ones defined here
  Pointer(_pSysKeyCount)   := @SysKeyCount;
  Pointer(_pSysKeyQue)     := @SysKeyQue;
  Pointer(_pSysShiftState) := @SysShiftState;
  Pointer(_pSysMouCount)   := @SysMouCount;
  Pointer(_pSysMouQue)     := @SysMouQue;

  // Determine the operating system (Win95/98 or Windows NT)
  OSVersionInfo.dwOSVersionInfoSize := SizeOf(OSVersionInfo);
  GetVersionEx(OSVersionInfo);
  SysPlatform := OSVersionInfo.dwPlatformId;
end;

{ Translate Windows CtrlKeys field to the low byte of OS/2-compatible
  ShiftStates, with the following valid values:
  Bit Value
   0  0001 Right Shift
   1  0002 Left shift
   2  0004 Either Ctrl
   3  0008 Either Alt
   4  0010 ScrollLock
   5  0020 NumLock
   6  0040 CapsLock
   --- unused:
   7  0080 Insert
   8  0100 Left Ctrl
   9  0200 Left Alt
  10  0400 Right Ctrl
  11  0800 Right Alt (AltGr)
  12  1000 ScrollLock down
  13  2000 NumLock down
  14  4000 CapsLock down
  15  8000 SysReq down
}

function CtrlKeysToShiftState(CtrlKeys: Integer): Byte;
begin
  Result := 0;
  if CtrlKeys and (RIGHT_ALT_PRESSED + LEFT_ALT_PRESSED) <> 0 then
    Result := $0008;
  if CtrlKeys and (RIGHT_CTRL_PRESSED + LEFT_CTRL_PRESSED) <> 0 then
    Inc(Result, $0004);
  if CtrlKeys and SHIFT_PRESSED <> 0 then
    Inc(Result, $0003);
  if CtrlKeys and SCROLLLOCK_ON <> 0 then
    Inc(Result, $0010);
  if CtrlKeys and NUMLOCK_ON <> 0 then
    Inc(Result, $0020);
  if CtrlKeys and CAPSLOCK_ON <> 0 then
    Inc(Result, $0040);
end;

type
  TVirtualKeyCodesArray = array[$00..$bf] of Byte;
  TShiftState = (Normal, Shift, Ctrl, Alt);

const
  VirtualKeyCodes: array[TShiftState] of TVirtualKeyCodesArray = (
  (
  //NRML x0|  x1|  x2|  x3|  x4|  x5|  x6|  x7|  x8|  x9|  xA|  xB|  xC|  xD|  xE|  xF|
  //0x |LMB |RMB |CBRK|MMB |X1MB|X2MB|----|BcSp|Tab |----|----|Clr |Entr|----|----|
          0,   0,   0,   0,   0,   0,   0,   0, $ff, $ff,   0,   0,  76, $ff,   0,   0,
  //1x Shft|Ctrl|Alt |Paus|Caps|IME |----|IME |IME |IME |----|Esc |IME |IME |IME |IME |
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0, $ff,   0,   0,   0,   0,
  //2x Spce|PgUp|PgDn|End |Home|Left|Up  |Rigt|Down|Slct|Prnt|Exec|SyRq|Ins |Del |Help|
        $ff,  73,  81,  79,  71,  75,  72,  77,  80,   0,   0,   0,   0,  82,  83,   0,
  //3x 0   |1   |2   |3   |4   |5   |6   |7   |8   |9   |----|----|----|----|----|----|
        $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff,   0,   0,   0,   0,   0,   0,
  //4x ----|A   |B   |C   |D   |E   |F   |G   |H   |I   |J   |K   |L   |M   |N   |O   |
          0, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff,
  //5x P   |Q   |R   |S   |T   |U   |V   |W   |X   |Y   |Z   |LWin|RWin|Prop|----|Slep|
        $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, 236, 237, 238,   0,   0,
  //6x num0|num1|num2|num3|num4|num5|num6|num7|num8|num9|Mult|Add |Sepa|Subt|Deci|Divi|
        $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff,
  //7x F1  |F2  |F3  |F4  |F5  |F6  |F7  |F8  |F9  |F10 |F11 |F12 |F13 |F14 |F15 |F16 |
         59,  60,  61,  62,  63,  64,  65,  66,  67,  68, 133, 134,   0,   0,   0,   0,
  //8x F17 |F18 |F19 |F20 |F21 |F22 |F23 |F24 |----|----|----|----|----|----|----|----|
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  //9x NuLk|ScLk|----|----|----|----|----|----|----|----|----|----|----|----|----|----|
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  //Ax LSft|RSft|LCtl|RCtl|LAlt|RAlt|bBak|bFwd|bRfs|bStp|bSch|bFav|bHme|bMte|vDwn|vUp |
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  //Bx mNxt|mPrv|mStp|mPly|mMal|mSel|mAp1|mAp2|----|----|Oem1|Oem+|Oem,|Oem-|Oem.|Oem2|
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0, $ff, $ff, $ff, $ff, $ff, $ff{,
  //Cx Oem3|----|----|----|----|----|----|----|----|----|----|----|----|----|----|----|
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  //Dx ----|----|----|----|----|----|----|----|----|----|----|Oem4|Oem5|Oem6|Oem7|Oem8|
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  //Ex OF17|OF18|O102|IHlp|Ico0|IME |IClr|----|----|ORst|OJmp|OPA1|OPA2|OPA3|OWSC|OCUS|
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  //Fx OAtt|OFin|OCpy|OAto|OENL|OBTb|Attn|CrSl|ExSl|EEof|Play|Zoom|NoNm|PA1 |OClr||
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0}
  ),
  (
  //SHFT x0|  x1|  x2|  x3|  x4|  x5|  x6|  x7|  x8|  x9|  xA|  xB|  xC|  xD|  xE|  xF|
  //0x |LMB |RMB |CBRK|MMB |X1MB|X2MB|----|BcSp|Tab |----|----|Clr |Entr|----|----|
          0,   0,   0,   0,   0,   0,   0,   0, $ff,  15,   0,   0,  76, $ff,   0,   0,
  //1x Shft|Ctrl|Alt |Paus|Caps|IME |----|IME |IME |IME |----|Esc |IME |IME |IME |IME |
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0, $ff,   0,   0,   0,   0,
  //2x Spce|PgUp|PgDn|End |Home|Left|Up  |Rigt|Down|Slct|Prnt|Exec|SyRq|Ins |Del |Help|
        $ff,  73,  81,  79,  71,  75,  72,  77,  80,   0,   0,   0,   0,  82,  83,   0,
  //3x 0   |1   |2   |3   |4   |5   |6   |7   |8   |9   |----|----|----|----|----|----|
        $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff,   0,   0,   0,   0,   0,   0,
  //4x ----|A   |B   |C   |D   |E   |F   |G   |H   |I   |J   |K   |L   |M   |N   |O   |
          0, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff,
  //5x P   |Q   |R   |S   |T   |U   |V   |W   |X   |Y   |Z   |LWin|RWin|Prop|----|Slep|
        $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, 236, 237, 238,   0,   0,
  //6x num0|num1|num2|num3|num4|num5|num6|num7|num8|num9|Mult|Add |Sepa|Subt|Deci|Divi|
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0, $ff,   0, $ff,   0, $ff,
  //7x F1  |F2  |F3  |F4  |F5  |F6  |F7  |F8  |F9  |F10 |F11 |F12 |F13 |F14 |F15 |F16 |
         84,  85,  86,  87,  88,  89,  90,  91,  92,  93, 135, 136,   0,   0,   0,   0,
  //8x F17 |F18 |F19 |F20 |F21 |F22 |F23 |F24 |----|----|----|----|----|----|----|----|
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  //9x NuLk|ScLk|----|----|----|----|----|----|----|----|----|----|----|----|----|----|
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  //Ax LSft|RSft|LCtl|RCtl|LAlt|RAlt|bBak|bFwd|bRfs|bStp|bSch|bFav|bHme|bMte|vDwn|vUp |
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  //Bx mNxt|mPrv|mStp|mPly|mMal|mSel|mAp1|mAp2|----|----|Oem1|Oem+|Oem,|Oem-|Oem.|Oem2|
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0, $ff, $ff, $ff, $ff, $ff, $ff
  ),
  (
  //CTRL x0|  x1|  x2|  x3|  x4|  x5|  x6|  x7|  x8|  x9|  xA|  xB|  xC|  xD|  xE|  xF|
  //0x |LMB |RMB |CBRK|MMB |X1MB|X2MB|----|BcSp|Tab |----|----|Clr |Entr|----|----|
          0,   0,   0,   0,   0,   0,   0,   0, $ff, 148,   0,   0,   0, $ff,   0,   0, // RETURN CLR???
  //1x Shft|Ctrl|Alt |Paus|Caps|IME |----|IME |IME |IME |----|Esc |IME |IME |IME |IME |
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  //2x Spce|PgUp|PgDn|End |Home|Left|Up  |Rigt|Down|Slct|Prnt|Exec|SyRq|Ins |Del |Help|
        $ff, 132, 118, 117, 119, 115, 141, 116, 145,   0,   0,   0,   0, 146,   6,   0, // $1D (Ctrl+Delete) fixed in code.
  //3x 0   |1   |2   |3   |4   |5   |6   |7   |8   |9   |----|----|----|----|----|----|
          0,   0,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0, // $36 (Ctrl+6) fixed in code.
  //4x ----|A   |B   |C   |D   |E   |F   |G   |H   |I   |J   |K   |L   |M   |N   |O   |
          0, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff,
  //5x P   |Q   |R   |S   |T   |U   |V   |W   |X   |Y   |Z   |LWin|RWin|Prop|----|Slep|
        $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, $ff, 236, 237, 238,   0,   0,
  //6x num0|num1|num2|num3|num4|num5|num6|num7|num8|num9|Mult|Add |Sepa|Subt|Deci|Divi|
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0, 114,   0,   0,   0,   0,   0,
  //7x F1  |F2  |F3  |F4  |F5  |F6  |F7  |F8  |F9  |F10 |F11 |F12 |F13 |F14 |F15 |F16 |
         94,  95,  96,  97,  98,  99, 100, 101, 102, 103, 137, 138,   0,   0,   0,   0,
  //8x F17 |F18 |F19 |F20 |F21 |F22 |F23 |F24 |----|----|----|----|----|----|----|----|
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  //9x NuLk|ScLk|----|----|----|----|----|----|----|----|----|----|----|----|----|----|
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  //Ax LSft|RSft|LCtl|RCtl|LAlt|RAlt|bBak|bFwd|bRfs|bStp|bSch|bFav|bHme|bMte|vDwn|vUp |
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  //Bx mNxt|mPrv|mStp|mPly|mMal|mSel|mAp1|mAp2|----|----|Oem1|Oem+|Oem,|Oem-|Oem.|Oem2|
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,  39,  13,   0,   0,   0,   0  // $BD (Ctrl+-) fixed in code.
  ),
  (
  //TODO: Alt+= should give 131, Alt+- should give 130
  //NOTE: Returns 1..9 (10=0) from numeric keypad for manual character code entry.
  //ALT  x0|  x1|  x2|  x3|  x4|  x5|  x6|  x7|  x8|  x9|  xA|  xB|  xC|  xD|  xE|  xF|
  //0x |LMB |RMB |CBRK|MMB |X1MB|X2MB|----|BcSp|Tab |----|----|Clr |Entr|----|----|
          0,   0,   0,   0,   0,   0,   0,   0,  14,   0,   0,   0,   0,   0,   0,   0,
  //1x Shft|Ctrl|Alt |Paus|Caps|IME |----|IME |IME |IME |----|Esc |IME |IME |IME |IME |
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  //2x Spce|PgUp|PgDn|End |Home|Left|Up  |Rigt|Down|Slct|Prnt|Exec|SyRq|Ins |Del |Help|
          0, 153, 161, 159, 151, 155, 152, 157, 160,   0,   0,   0,   0, 162, 163,   0,
  //3x 0   |1   |2   |3   |4   |5   |6   |7   |8   |9   |----|----|----|----|----|----|
        129, 120, 121, 122, 123, 124, 125, 126, 127, 128,   0,   0,   0,   0,   0,   0,
  //4x ----|A   |B   |C   |D   |E   |F   |G   |H   |I   |J   |K   |L   |M   |N   |O   |
          0,  30,  48,  46,  32,  18,  33,  34,  35,  23,  36,  37,  38,  50,  49,  24,
  //5x P   |Q   |R   |S   |T   |U   |V   |W   |X   |Y   |Z   |LWin|RWin|Prop|----|Slep|
         25,  16,  19,  31,  20,  22,  47,  17,  45,  21,  44, 236, 237, 238,   0,   0,
  //6x num0|num1|num2|num3|num4|num5|num6|num7|num8|num9|Mult|Add |Sepa|Subt|Deci|Divi|
         10,   1,   2,   3,   4,   5,   6,   7,   8,   9,   0,   0,   0,   0,   0,   0,
  //7x F1  |F2  |F3  |F4  |F5  |F6  |F7  |F8  |F9  |F10 |F11 |F12 |F13 |F14 |F15 |F16 |
        104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 139, 140,   0,   0,   0,   0,
  //8x F17 |F18 |F19 |F20 |F21 |F22 |F23 |F24 |----|----|----|----|----|----|----|----|
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  //9x NuLk|ScLk|----|----|----|----|----|----|----|----|----|----|----|----|----|----|
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  //Ax LSft|RSft|LCtl|RCtl|LAlt|RAlt|bBak|bFwd|bRfs|bStp|bSch|bFav|bHme|bMte|vDwn|vUp |
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  //Bx mNxt|mPrv|mStp|mPly|mMal|mSel|mAp1|mAp2|----|----|Oem1|Oem+|Oem,|Oem-|Oem.|Oem2|
          0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0, 131,   0, 130,   0,   0
  ));

procedure KbdUpdateEventQueues;
const
  AltNumeric: Byte = 0;
  Platform: Longint = -1;
var
  EventCount: DWord;
  InRec: TInputRecord;
  FoundAlt: Boolean;
  ShiftState: TShiftState;
begin
  if SysKeyCount > High(SysKeyQue) then
    exit;
  repeat
    GetNumberOfConsoleInputEvents(SysFileStdIn, EventCount);
    if EventCount = 0 then
      Exit;
    ReadConsoleInput(SysFileStdIn, InRec, 1, EventCount);
    with InRec do
      case EventType of
        key_Event:
          if SysKeyCount <= High(SysKeyQue) then
          with SysKeyQue[SysKeyCount], KeyEvent do
          begin
            skeShiftState := CtrlKeysToShiftState(dwControlKeyState);
            SysShiftState := skeShiftState;
            if bKeyDown then
              begin
                // Figure out shift status to assume during the rest of the processing, prioritized like in BP.
                if ((dwControlKeyState and (LEFT_ALT_PRESSED or RIGHT_ALT_PRESSED)) <> 0) then
                  ShiftState := Alt
                else if ((dwControlKeyState and (LEFT_CTRL_PRESSED or RIGHT_CTRL_PRESSED)) <> 0) then
                  ShiftState := Ctrl
                else if ((dwControlKeyState and SHIFT_PRESSED) <> 0) then
                  ShiftState := Shift
                else
                  ShiftState := Normal;

                // Lookup virtual key code in table, if code is too large, use $ff (= return AsciiChar unchanged).
                if (wVirtualKeyCode > High(VirtualKeyCodes[ShiftState])) then
                  skeKeyCode := $ff
                else
                  skeKeyCode := VirtualKeyCodes[ShiftState, wVirtualKeyCode];

                // Figure out if we are doing Alt+numpad character code entry.
                FoundAlt := (ShiftState = Alt) and (skeKeyCode in [1..10]);

                // Assume some special processing of the Alt shift state, this is for character code entry and Alt vs. AltGr support.
                if (ShiftState = Alt) then
                  if FoundAlt then
                    if (skeKeyCode = 10) then
                      AltNumeric := (AltNumeric * 10)
                    else
                      AltNumeric := (AltNumeric * 10) + skeKeyCode
                  else if ((dwControlKeyState and RIGHT_ALT_PRESSED) <> 0) or ((dwControlKeyState and (LEFT_CTRL_PRESSED or RIGHT_CTRL_PRESSED)) <> 0) then
                    if (wVirtualKeyCode in [$30..$39, $41..$5A]) then
                      if not (wVirtualKeyCode in [Ord(AsciiChar), (Ord(AsciiChar) - (Ord('a') - Ord('A')))]) then
                        skeKeyCode := $ff;

                // We are not doing Alt + numpad character code input so process and return result.
                if not FoundAlt then
                begin

                  if (skeKeyCode = $ff) then
                    skeKeyCode := (wVirtualScanCode shl 8) or Ord(AsciiChar)
                  else if (skeKeyCode <> 0) then
                    skeKeyCode := skeKeyCode shl 8
                  else if (ShiftState = Ctrl) then
                    case wVirtualKeyCode of
                      $2e: skeKeyCode := $ff00; // Delete
                      $36: skeKeyCode := $001e; // Ctrl+6
                      $bd: skeKeyCode := $001f; // Ctrl+-
                    end; // case wVirtualKeyCode of

                  if (skeKeyCode = 0) then
                    Exit;

                  Inc(SysKeyCount);
                  AltNumeric := 0;
                end; // if not FoundAlt then

              end // if bKeyDown then
            else
              if (KeyEvent.wVirtualKeyCode = VK_MENU) then
              if (AltNumeric <> 0) then
                begin
                  skeKeyCode := AltNumeric;
                  AltNumeric := 0;
                  inc(SysKeyCount);
                end;
          end;
        _mouse_Event:
          if SysMouCount <= High(SysMouQue) then
          with SysMouQue[SysMouCount] do
          begin
            smePos.X := MouseEvent.dwMousePosition.X;
            smePos.Y := MouseEvent.dwMousePosition.Y;
            smeButtons := MouseEvent.dwButtonState and $0003;
            smeTime := SysSysMsCount;
            Inc(SysMouCount);
          end;
      end;
    until (False) OR (SysKeyCount > High(SysKeyQue));
end;

{$Ifdef KeyDll}
Exports
  KbdInit              name 'KbdInit',
  KbdUpdateEventQueues name 'KbdUpdateEventQueues';
{$Endif}

begin
end.
